﻿using System.Configuration;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;

[ExecuteInEditMode()]
public class TextureBlendAnimation : MonoBehaviour {
    public List<Texture2D> textures  = new List<Texture2D>();
    private List<float> texLevels = new List<float>();

    private Camera cam;
    private Material texMat;
    
    public RenderTexture targetTexture;
    private bool hasTargetTexture = false;

    [Range(0, 1)]
    public float blendPosition = 0;

    private float prevBlendPosition = -1;

    private RenderTexture rendTex;
    private bool hasRendTex = false;

    [Range(2, 4096)]
    public int texWidth = 256;

    [Range(2, 4096)]
    public int texHeight = 256;

    void LateUpdate()
    {
        if (textures.Count != texLevels.Count)
        {
            texLevels.Clear();
            texLevels.AddRange(new float[textures.Count]);
        }

        if (hasRendTex && !hasTargetTexture)
        {
            if (rendTex.width != texWidth | rendTex.height != texHeight)
            {
                if (rendTex)
                {
                    cam.targetTexture = null;
                    DestroyImmediate(rendTex);
                    rendTex = new RenderTexture(texWidth, texHeight, 0);
                    cam.targetTexture = rendTex;
                }
            }
        }

        if (prevBlendPosition != blendPosition)
        {
            prevBlendPosition = blendPosition;

            float tl = 1.0f / (float)textures.Count;
            for (int i = 0; i < textures.Count; i++)
            {
                float cmin = tl*(i - 1);
                float cmax = tl*(i);
                texLevels[i] = Unity.PerForm.Utils.Misc.MapValue(blendPosition, cmin + tl * 0.5f, cmax + tl * 0.5f, 0, 2);
                if (i > 0)
                {
                    if (texLevels[i] > 1)
                    {
                        texLevels[i - 1] = 2 - texLevels[i];
                        texLevels[i] = 1;
                    }
                }
                else
                {
                    if (texLevels[i] > 1)
                    {
                        texLevels[i] = 1;
                    }
                }
            }
        }

#if UNITY_EDITOR
        if (Application.isEditor && !Application.isPlaying)
        {
            hasTargetTexture = targetTexture != null;
        }
#endif
    }

    void Start()
    {
        texMat = new Material(Resources.Load<Shader>("Shaders/TextureBlend"));

        for (int i = 0; i < textures.Count; i++)
        {
            texLevels.Add(0);
        }

        hasTargetTexture = targetTexture != null;
        if (hasTargetTexture)
        {
            rendTex = targetTexture;
            texWidth = rendTex.width;
            texHeight = rendTex.height;
        }
        else
        {
            rendTex = new RenderTexture(texWidth, texHeight, 0);
        }

        rendTex.filterMode = FilterMode.Trilinear;

        cam = gameObject.AddComponent<Camera>();
        cam.clearFlags = CameraClearFlags.Color;
        cam.backgroundColor = new  Color(0, 0, 0, 0);
        cam.orthographic = true;
        cam.nearClipPlane = 0.1f;
        cam.farClipPlane = 0.2f;
        cam.orthographicSize = 0.1f;
        cam.useOcclusionCulling = false;
        cam.renderingPath = RenderingPath.Forward;
        cam.targetTexture = rendTex;
        cam.cullingMask = 0;
        cam.hideFlags = HideFlags.HideAndDontSave;

        hasRendTex = true;

    }

    private bool isDisableing = false;

    void OnDisabled()
    {
        if (Application.isEditor)
        {
            isDisableing = true;
            OnDestroy();
            isDisableing = false;
        }
    }

    void OnDestroy()
    {
        if (isDisableing)
            return;

        if (texMat != null)
        {            
            //texMat

            if (Application.isEditor)
            {
                DestroyImmediate(texMat);                
            } else 
            {
                Destroy(texMat);            
            }
            texMat = null;
        }


        if (!hasTargetTexture)
        {
            if (rendTex != null)
            {
                cam.targetTexture = null;
                if (Application.isEditor)
                {
                    DestroyImmediate(rendTex);
                }
                else
                {
                    //Destroy(rendTex);                    
                }
                rendTex = null;
            }
        }

        if (cam != null)
        {
            if (cam.targetTexture != null)
            {
                cam.targetTexture = null;                
            }
            //if (cam != null)
                //DestroyImmediate(cam);

            cam = null;
        }

        if (texMat)
        {
            if (Application.isEditor)
            {
                DestroyImmediate(texMat);
            }
            else
            {
                Destroy(texMat);                
            }
        }

    }

//#if UNITY_EDITOR
//    void OnGUI()
//    {
//        string levelsInfo = "";
//        for (int ti = 0; ti < textures.Count; ti++)
//        {
//            var lev = texLevels[ti];
//            levelsInfo += ti.ToString("00") + ": " + lev.ToString("n2") + "\n";
//        }

//        GUI.Label(new Rect(10f, 50f, 100f, 50f), levelsInfo);
//        //Debug.Log(levelsInfo);
//    }
//#endif

    private void OnPostRender()
    {
        if (!hasRendTex)
            return;


        for (int ti = 0; ti < textures.Count; ti++)
        {

            Texture2D tex = textures[ti];
            if (tex != null)
            {
                if (texLevels[ti] > 0)
                {
                    
                }
                GL.LoadOrtho();
                GL.Viewport(new Rect(0, 0, rendTex.width, rendTex.height));
                texMat.SetTexture("_CamTex", tex);
                texMat.SetFloat("_Alpha", texLevels[ti]);
                texMat.SetPass(0);
                GL.Begin(GL.QUADS);
                GL.TexCoord2(0, 0);
                GL.Vertex3(0.0F, 0.0F, 0);
                GL.TexCoord2(0, 1);
                GL.Vertex3(0.0F, 1.0F, 0);
                GL.TexCoord2(1, 1);
                GL.Vertex3(1.0F, 1.0F, 0);
                GL.TexCoord2(1, 0);
                GL.Vertex3(1.0F, 0.0F, 0);
                GL.End();
            }
        }
        

        GL.PushMatrix();
        GL.LoadOrtho();

        GL.PopMatrix();
    }
}
